"""
OAuth2 provider customized `django-oauth2-provider` forms.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import logging

from django.contrib.auth import authenticate
from django.contrib.auth.models import User
import provider.constants
import provider.oauth2.forms
from provider.forms import OAuthValidationError
from provider.oauth2.forms import ScopeChoiceField
from provider.oauth2.models import Client

from .constants import SCOPE_NAMES

log = logging.getLogger(__name__)


# The following forms override the scope field to use the SCOPE_NAMES
# defined for this provider. Otherwise it will use the default values from
# the django-oauth2-provider package.

# pylint: disable=missing-docstring,no-member

class AuthorizationRequestForm(provider.oauth2.forms.AuthorizationRequestForm):
    def __init__(self, *args, **kwargs):
        super(AuthorizationRequestForm, self).__init__(*args, **kwargs)
        self.fields['scope'] = ScopeChoiceField(choices=SCOPE_NAMES, required=False)


class AuthorizationForm(provider.oauth2.forms.AuthorizationForm):
    def __init__(self, *args, **kwargs):
        super(AuthorizationForm, self).__init__(*args, **kwargs)
        self.fields['scope'] = ScopeChoiceField(choices=SCOPE_NAMES, required=False)


class RefreshTokenGrantForm(provider.oauth2.forms.RefreshTokenGrantForm):
    def __init__(self, *args, **kwargs):
        super(RefreshTokenGrantForm, self).__init__(*args, **kwargs)
        self.fields['scope'] = ScopeChoiceField(choices=SCOPE_NAMES, required=False)


class AuthorizationCodeGrantForm(provider.oauth2.forms.AuthorizationCodeGrantForm):
    def __init__(self, *args, **kwargs):
        super(AuthorizationCodeGrantForm, self).__init__(*args, **kwargs)
        self.fields['scope'] = ScopeChoiceField(choices=SCOPE_NAMES, required=False)


# pylint: enable=missing-docstring,no-member

# The forms in this module are required to use email as a secondary
# identifier when authenticating via OAuth2, since the specification
# only uses the `username` parameter.
#
# TODO: An alternative, simpler, approach is to write a backend that
# like `django.contrib.auth.backends.ModelBackend` and add it to
# `AUTHENTICATION_BACKENDS` in the Django settings.


class PasswordGrantForm(provider.oauth2.forms.PasswordGrantForm):
    """
    Forms that validates the user email to be used as secondary user
    identifier during authentication.
    """
    def clean(self):
        data = self.cleaned_data  # pylint: disable=no-member
        username = data.get('username')
        password = data.get('password')

        user = authenticate(username=username, password=password)

        # If the username was not found try the user using username as
        # the email address. It is valid because the edx-platform has
        # a unique constraint placed on the email field.
        if user is None:
            try:
                user_obj = User.objects.get(email=username)
                user = authenticate(username=user_obj.username, password=password)
            except User.DoesNotExist:
                user = None

        if user is None:
            # TODO This is a temporary workaround while the is_active field on the
            # user is coupled with whether or not the user has verified ownership
            # of their claimed email address.  Once is_active is decoupled from
            # verified_email, add the following condition to the 'if' statement above.
            # or not user.is_active
            error_description = "Username does not exist or invalid credentials given for username '{}'.".format(
                username
            )
            log.error("OAuth2: {}".format(error_description))
            raise OAuthValidationError({
                'error': 'invalid_grant',
                'error_description': error_description
            })

        data['user'] = user
        return data


class PublicPasswordGrantForm(PasswordGrantForm, provider.oauth2.forms.PublicPasswordGrantForm):
    """
    Form wrapper to ensure the the customized PasswordGrantForm is used
    during client authentication.
    """
    def clean(self):
        data = super(PublicPasswordGrantForm, self).clean()

        try:
            client = Client.objects.get(client_id=data.get('client_id'))
        except Client.DoesNotExist:
            error_description = "Client ID '{}' does not exist.".format(data.get('client_id'))
            log.exception("OAuth2: {}".format(error_description))
            raise OAuthValidationError({
                'error': 'invalid_client',
                'error_description': error_description
            })

        if client.client_type != provider.constants.PUBLIC:
            error_description = "'{}' is not a public client.".format(client.client_type)
            log.error("OAuth2: {}".format(error_description))
            raise OAuthValidationError({
                'error': 'invalid_client',
                'error_description': error_description
            })
        data['client'] = client
        return data



# pylint: disable=missing-docstring

from django.test import TestCase
from provider.oauth2.models import AccessToken, Client, RefreshToken

from .factories import AccessTokenFactory, ClientFactory, RefreshTokenFactory, UserFactory


class TestFactory(TestCase):

    def test_client_factory(self):
        user = UserFactory()
        actual_client = ClientFactory(user=user)
        expected_client = Client.objects.get(user=user)
        self.assertEqual(actual_client, expected_client)

    def test_access_token_client_factory(self):
        user = UserFactory()
        actual_access_token = AccessTokenFactory(user=user, client=ClientFactory())
        expected_access_token = AccessToken.objects.get(user=user)
        self.assertEqual(actual_access_token, expected_access_token)

    def test_refresh_token_factory(self):
        user = UserFactory()
        client = ClientFactory()
        access_token = AccessTokenFactory(user=user, client=client)
        actual_refresh_token = RefreshTokenFactory(user=user, client=client, access_token=access_token)
        expected_refresh_token = RefreshToken.objects.get(user=user, access_token=access_token)
        self.assertEqual(actual_refresh_token, expected_refresh_token)



"""
Tests of the create_oauth2_client management command.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import json
from itertools import product

import ddt
from django.contrib.auth import get_user_model
from django.core.management import call_command
from django.core.management.base import CommandError
from django.test import TestCase
from django.utils.six import StringIO
from provider.oauth2.models import Client

from ..models import TrustedClient

URL = 'https://www.example.com/'
REDIRECT_URI = 'https://www.example.com/complete/edx-oidc/'
CLIENT_TYPES = ('confidential', 'public')
USERNAME = 'username'


@ddt.ddt
class CreateOauth2ClientTests(TestCase):
    """
    Mgmt cmd test class.
    """
    def setUp(self):
        super(CreateOauth2ClientTests, self).setUp()
        user_model = get_user_model()
        self.user = user_model.objects.create(username=USERNAME)

    def _call_command(self, args, options=None):
        """Call the command, capturing its output."""
        if options is None:
            options = {}

        out = StringIO()
        options['stdout'] = out

        call_command('create_oauth2_client', *args, **options)

        return out

    def assert_client_created(self, args, options):
        """Call the command, verify that the Client was created, and validate the output."""
        out = self._call_command(args, options)
        client = Client.objects.get()

        # Verify command output
        serialized = json.dumps(client.serialize(), indent=4)
        self.assertIn(serialized, out.getvalue())

        # Verify Client associated with the correct user
        if options.get('username'):
            self.assertEqual(self.user, client.user)

        # Verify Client assigned the correct name
        client_name = options.get('client_name')
        if client_name:
            self.assertEqual(client_name, client.name)

        # Verify Client ID and secret overrides
        for attr in ('client_id', 'client_secret'):
            value = options.get(attr)
            if value is not None:
                self.assertEqual(value, getattr(client, attr))

        # Verify Client designated as trusted
        if options.get('trusted'):
            trusted_client = TrustedClient.objects.get()
            self.assertEqual(client, trusted_client.client)

    # Generate all valid argument and options combinations
    @ddt.data(*product(
        # Generate all valid argument combinations
        product(
            (URL,),
            (REDIRECT_URI,),
            (t for t in CLIENT_TYPES),
        ),
        # Generate all valid option combinations
        (dict(zip(('username', 'client_name', 'client_id', 'client_secret', 'trusted'), p)) for p in product(
            (USERNAME, None),
            ('name', None),
            ('id', None),
            ('secret', None),
            (True, False)
        ))
    ))
    @ddt.unpack
    def test_client_creation(self, args, options):
        """Verify that the command creates a Client when given valid arguments and options."""
        self.assert_client_created(args, options)

    @ddt.data(
        ((URL, REDIRECT_URI), 'too few arguments'),
        ((URL, REDIRECT_URI, CLIENT_TYPES[0], CLIENT_TYPES[1]), 'unrecognized arguments'),
    )
    @ddt.unpack
    def test_argument_cardinality(self, args, err_msg):
        """Verify that the command fails when given an incorrect number of arguments."""
        with self.assertRaises(CommandError) as exc:
            self._call_command(args, {})

        self.assertIn(err_msg, exc.exception.message)

    @ddt.data(
        ('invalid', REDIRECT_URI, CLIENT_TYPES[0]),
        (URL, 'invalid', CLIENT_TYPES[0]),
    )
    def test_url_validation(self, args):
        """Verify that the command fails when the provided URLs are invalid."""
        with self.assertRaises(CommandError) as exc:
            self._call_command(args)

        self.assertIn('URLs provided are invalid.', exc.exception.message)

    def test_client_type_validation(self):
        """Verify that the command fails when the provided client type is invalid."""
        with self.assertRaises(CommandError) as exc:
            self._call_command((URL, REDIRECT_URI, 'not_a_client_type'))

        self.assertIn('Client type provided is invalid.', exc.exception.message)

    def test_username_mismatch(self):
        """Verify that the command fails when the provided username is invalid."""
        with self.assertRaises(CommandError) as exc:
            self._call_command(
                (URL, REDIRECT_URI, CLIENT_TYPES[0]),
                options={'username': 'bad_username'}
            )

        self.assertIn('User matching the provided username does not exist.', exc.exception.message)

    def test_idempotency(self):
        """Verify that the command can be run repeatedly with the same client ID, without any ill effects."""
        args = [URL, REDIRECT_URI, CLIENT_TYPES[0]]
        options = {
            'username': None,
            'client_name': 'name',
            'client_id': 'id',
            'client_secret': 'secret',
            'trusted': True
        }

        self.assert_client_created(args, options)

        # Verify that the command is idempotent.
        self.assert_client_created(args, options)

        # Verify that attributes are updated if the command is run with the same client ID,
        # but with other options varying.
        options['client_secret'] = 'another-secret'
        self.assert_client_created(args, options)

        # Verify that the TrustedClient is deleted if the command is run with the same
        # client ID, but with the "--trusted" flag excluded.
        options['trusted'] = False
        self.assert_client_created(args, options)
        self.assertEqual(len(TrustedClient.objects.all()), 0)



"""
Default OpenID Connect claim handlers.

Scopes are groups of claims. Claims are simple assertions about a
OpenID Connect user or client. For more information see:

http://openid.net/specs/openid-connect-core-1_0.html#Claims

Claim handlers are simple python objects that serve the following
purposes:

  - Determine what scopes and claims are available to each specific
    user and client.
  - Determine what claims are associated with each scope.
  - Evaluate the values of claims for each user and client.

Claim handlers do this by defining two different type of methods, one
corresponding to scope, and the other one corresponding to claims.

Scope Methods

The scope methods start with the prefix `scope_` and end with the name
of the scope, for example `scope_profile`. The method receives a
single parameter `data`, a dictionary with the following fields:

  - 'user': Requied. User instance for the current request.
  - 'client': Required. OAuth2 Client instance for the current request.

Each scope method should return a list of its asociated claim names,
or None if the `user` or `client` don't have authorization for that
scope.

Claim Methods

The claim methods start with the prefix `claim_` and end with the name
of the claim, for example `claim_email`.  The method receives a
single parameter `data`, a dictionary with the following fields:

  - 'user': Requied. User instance for the current request.
  - 'client': Required. OAuth2 Client instance for the current request.
  - 'value': Optional. Request that the claim returns with a particular value.
  - 'values': Optional. Request that the claim returns with particular values.
  - 'essential': Optional.  Indicate if the claim is essential.

For information on the 'value', 'values' and 'essential' fields, and
other claim values, see:

http://openid.net/specs/openid-connect-core-1_0.html#IndividualClaimsRequests

Claim methods return the value corresponding to the claim according to
the passed data fields. The interpretetion of the value(s) field is open to
each particular claim, and should be documented. If the returned value
is `None`, the claim will not be included in the response.


NOTE: The method `__getattr__` can be overloaded to support claims or
scopes whose names are not valid python method names.

"""

# pylint: disable=unused-argument
from __future__ import absolute_import, division, print_function, unicode_literals

from calendar import timegm
from datetime import datetime, timedelta

from django.conf import settings


class BasicIDTokenHandler(object):
    """
    Basic OpenID Connect ID token claims.

    For reference see:
    http://openid.net/specs/openid-connect-basic-1_0.html#IDToken

    """

    def __init__(self):
        self._now = None

    @property
    def now(self):
        """ Capture time. """
        if self._now is None:
            # Compute the current time only once per instance
            self._now = datetime.utcnow()
        return self._now

    def scope_openid(self, data):
        """ Returns claims for the `openid` profile. """
        return ['iss', 'sub', 'aud', 'iat', 'exp', 'nonce']

    def claim_iss(self, data):
        """ Required issuer identifier. """
        return settings.OAUTH_OIDC_ISSUER

    def claim_sub(self, data):
        """ Required subject identifier. """
        # Use the primary key as the identifier
        return str(data['user'].pk)

    def claim_aud(self, data):
        """ Required audience. """
        return data['client'] .client_id

    def claim_iat(self, data):
        """ Required current/issued time. """
        return timegm(self.now.utctimetuple())

    def claim_exp(self, data):
        """ Required expiration time. """
        expiration = getattr(settings, 'OAUTH_ID_TOKEN_EXPIRATION', 30)
        expires = self.now + timedelta(seconds=expiration)
        return timegm(expires.utctimetuple())

    def claim_nonce(self, data):
        """ Optional replay attack protection. """
        return data.get('value')


class BasicUserInfoHandler(object):
    """
    Basic OpenID Connect User Info claims.

    For reference see:
    http://openid.net/specs/openid-connect-basic-1_0.html#UserInfo

    """

    def scope_openid(self, data):
        """Returns claims for the `openid` profile"""
        return ['sub']

    def claim_sub(self, data):
        """ Required subject identifier. """
        # Use the primary key as the identifier
        return str(data['user'].pk)


class ProfileHandler(object):
    """
    Basic profile scope claim handler.

    For information see:
    http://openid.net/specs/openid-connect-basic-1_0.html#Scopes

    Some claims are missing since they require a detailed User
    profile, and can included from another claim handler.

    """

    def scope_profile(self, data):
        """ Returns claims for the `profile` scope. """
        return ['name', 'family_name', 'given_name', 'preferred_username']

    def claim_family_name(self, data):
        """ End user last or family name. """
        return data['user'].last_name

    def claim_name(self, data):
        """ End user full name. """
        return data['user'].get_full_name()

    def claim_given_name(self, data):
        """ End user first or given name. """
        return data['user'].first_name

    def claim_preferred_username(self, data):
        """ End user preferred username. """
        return data['user'].username


class EmailHandler(object):
    """
    Basic email scope claim handler.

    For information see:
    http://openid.net/specs/openid-connect-basic-1_0.html#Scopes

    Some claims are missing since they require a more detailed User profile.

    """

    def scope_email(self, data):
        """ Returns claims for the `profile` scope. """
        return ['email']

    def claim_email(self, data):
        """ End user email. """
        return data['user'].email



"""
Customized django-oauth2-provider views, aligned with the OpenID specification.

"""
from __future__ import absolute_import, division, print_function, unicode_literals

import json

from django.http import HttpResponse
from django.views.generic import View
import provider.oauth2.forms
import provider.oauth2.views
import provider.scope
from provider.oauth2.models import AccessToken
from provider.oauth2.views import Capture, OAuthError, Redirect  # pylint: disable=unused-import

from . import constants, oidc
from .backends import PublicPasswordBackend
from .forms import (
    AuthorizationCodeGrantForm,
    AuthorizationForm,
    AuthorizationRequestForm,
    PasswordGrantForm,
    RefreshTokenGrantForm
)
from .models import TrustedClient


# pylint: disable=abstract-method
class Authorize(provider.oauth2.views.Authorize):
    """
    edX customized authorization view:
      - Introduces trusted clients, which do not require user consent.
    """
    def get_request_form(self, client, data):
        return AuthorizationRequestForm(data, client=client)

    def get_authorization_form(self, _request, client, data, client_data):
        # Check if the client is trusted. If so, bypass user
        # authorization by filling the data in the form.
        trusted = TrustedClient.objects.filter(client=client).exists()
        if trusted:
            scope_names = provider.scope.to_names(client_data['scope'])
            data = {'authorize': ['Authorize'], 'scope': scope_names, 'nonce': client_data.get('nonce', '')}

        form = AuthorizationForm(data)
        return form

    def handle(self, request, post_data=None):
        response = super(Authorize, self).handle(request, post_data)

        if response.status_code < 400:
            # Store the ID of the client being used for authorization. We will use
            # this later to determine which clients to log out.
            client_id = request.session.get('oauth:client', {}).get('client_id')

            if client_id:
                client_ids = set(request.session.get(constants.AUTHORIZED_CLIENTS_SESSION_KEY, []))
                client_ids.add(client_id)
                request.session[constants.AUTHORIZED_CLIENTS_SESSION_KEY] = list(client_ids)

        return response


# pylint: disable=abstract-method
class AccessTokenView(provider.oauth2.views.AccessTokenView):
    """
    Customized OAuth2 access token view.

    Allows usage of email as main identifier when requesting a password grant.

    Support the ID Token endpoint following the OpenID Connect specification:

    - http://openid.net/specs/openid-connect-core-1_0.html#TokenEndpoint

    By default it returns all the claims available to the scope requested,
    and available to the claim handlers configured by `OAUTH_OIDC_ID_TOKEN_HANDLERS`

    """

    # Add custom authentication provider, to support email as username.
    authentication = (provider.oauth2.views.AccessTokenView.authentication +
                      (PublicPasswordBackend, ))

    # The following grant overrides make sure the view uses our customized forms.

    # pylint: disable=no-member
    def get_authorization_code_grant(self, _request, data, client):
        form = AuthorizationCodeGrantForm(data, client=client)
        if not form.is_valid():
            raise OAuthError(form.errors)
        return form.cleaned_data.get('grant')

    # pylint: disable=no-member
    def get_refresh_token_grant(self, _request, data, client):
        form = RefreshTokenGrantForm(data, client=client)
        if not form.is_valid():
            raise OAuthError(form.errors)
        return form.cleaned_data.get('refresh_token')

    # pylint: disable=no-member
    def get_password_grant(self, _request, data, client):
        # Use customized form to allow use of user email during authentication.
        form = PasswordGrantForm(data, client=client)
        if not form.is_valid():
            raise OAuthError(form.errors)
        return form.cleaned_data

    # pylint: disable=super-on-old-class
    def access_token_response_data(self, access_token, response_type=None, nonce=''):
        """
        Return `access_token` fields for OAuth2, and add `id_token` fields for
        OpenID Connect according to the `access_token` scope.

        """

        # Clear the scope for requests that do not use OpenID Connect.
        # Scopes for pure OAuth2 request are currently not supported.
        scope = constants.DEFAULT_SCOPE

        extra_data = {}

        # Add OpenID Connect `id_token` if requested.
        #
        # TODO: Unfourtunately because of how django-oauth2-provider implements
        # scopes, we cannot check if `openid` is the first scope to be
        # requested, as required by OpenID Connect specification.

        if provider.scope.check(constants.OPEN_ID_SCOPE, access_token.scope):
            id_token = self.get_id_token(access_token, nonce)
            extra_data['id_token'] = self.encode_id_token(id_token)
            scope = provider.scope.to_int(*id_token.scopes)

        # Update the token scope, so it includes only authorized values.
        access_token.scope = scope
        access_token.save()

        # Get the main fields for OAuth2 response.
        response_data = super(AccessTokenView, self).access_token_response_data(access_token)

        # Add any additional fields if OpenID Connect is requested. The order of
        # the addition makes sures the OAuth2 values are not overrided.
        response_data = dict(extra_data.items() + response_data.items())

        return response_data

    def get_id_token(self, access_token, nonce):
        """ Return an ID token for the given Access Token. """

        claims_string = self.request.POST.get('claims')
        claims_request = json.loads(claims_string) if claims_string else {}

        return oidc.id_token(access_token, nonce, claims_request)

    def encode_id_token(self, id_token):
        """
        Return encoded ID token.

        """

        # Encode the ID token using the `client_secret`.
        #
        # TODO: Using the `client_secret` is not ideal, since it is transmitted
        # over the wire in some authentication flows.  A better alternative is
        # to use the public key of the issuer, which also allows the ID token to
        # be shared among clients. Doing so however adds some operational
        # costs. We should consider this for the future.

        secret = id_token.access_token.client.client_secret

        return id_token.encode(secret)


class ProtectedView(View):
    """
    Base View that checks for a valid OAuth2 access token.

    """

    # TODO: convert to a decorator.

    access_token = None
    user = None

    def dispatch(self, request, *args, **kwargs):
        error_msg = None

        # Get the header value
        token = request.META.get('HTTP_AUTHORIZATION', '')

        # Trim the Bearer portion
        token = token.replace('Bearer ', '')

        if token:
            # Verify token exists and is valid
            access_token = AccessToken.objects.filter(token=token)
            if access_token:
                access_token = access_token[0]
            else:
                access_token = None

            if access_token is None or access_token.get_expire_delta() <= 0:
                error_msg = 'invalid_token'
            else:
                self.access_token = access_token
                self.user = access_token.user
        else:
            # Return an error response if no token supplied
            error_msg = 'access_denied'

        if error_msg:
            return JsonResponse({'error': error_msg}, status=401)

        return super(ProtectedView, self).dispatch(request, *args, **kwargs)


class UserInfoView(ProtectedView):
    """
    Implementation of the Basic OpenID Connect UserInfo endpoint as described in:

    - http://openid.net/specs/openid-connect-basic-1_0.html#UserInfo

    By default it returns all the claims available to the `access_token` used, and available
    to the claim handlers configured by `OAUTH_OIDC_USERINFO_HANDLERS`

    In addition to the standard UserInfo response, this view also accepts custom scope
    and claims requests, using the scope and claims parameters as described in:

    http://openid.net/specs/openid-connect-core-1_0.html#ScopeClaims
    http://openid.net/specs/openid-connect-core-1_0.html#ClaimsParameter

    Normally, such requests can only be done when requesting an ID Token. However, it is
    also convinient to support then in the UserInfo endpoint to do simply authorization checks.

    It ignores the top level claims request for `id_token`, in the claims
    request, using only the `userinfo` section.

    All requests to this endpoint must include at least the 'openid' scope.

    Currently only supports GET request, and does not sign any responses.

    """

    def get(self, request, *_args, **_kwargs):
        """
        Respond to a UserInfo request.

        Two optional query parameters are accepted, scope and claims.
        See the references above for more details.

        """

        access_token = self.access_token

        scope_string = request.GET.get('scope')
        scope_request = scope_string.split() if scope_string else None

        claims_string = request.GET.get('claims')
        claims_request = json.loads(claims_string) if claims_string else None

        if not provider.scope.check(constants.OPEN_ID_SCOPE, access_token.scope):
            return self._bad_request('Missing openid scope.')

        try:
            claims = self.userinfo_claims(access_token, scope_request, claims_request)
        except ValueError, exception:
            return self._bad_request(str(exception))

        # TODO: Encode and sign responses if requested.

        response = JsonResponse(claims)

        return response

    def userinfo_claims(self, access_token, scope_request, claims_request):
        """ Return the claims for the requested parameters. """
        id_token = oidc.userinfo(access_token, scope_request, claims_request)
        return id_token.claims

    def _bad_request(self, msg):
        """ Return a 400 error with JSON content. """
        return JsonResponse({'error': msg}, status=400)


class JsonResponse(HttpResponse):
    """ Simple JSON Response wrapper. """
    def __init__(self, content, status=None, content_type='application/json'):
        super(JsonResponse, self).__init__(
            content=json.dumps(content),
            status=status,
            content_type=content_type,
        )



""" Test email as username authentication """
from __future__ import absolute_import, division, print_function, unicode_literals

import json

import ddt
from django.core.urlresolvers import reverse
from provider.constants import CONFIDENTIAL, PUBLIC

from .base import OAuth2TestCase
from .factories import ClientFactory

USERNAME = 'some_username'
EMAIL = 'test@example.com'
PASSWORD = 'some_password'

CLIENT_ID = 'some_client_id'
CLIENT_SECRET = 'some_secret'

# Data to generate login tests. Missing fields default to the values
# in the class variables below. 'success' defaults to False.
# 'client_secret' is not used unless specified.
AUTHENTICATION_TEST_DATA = [
    {
        'success': True
    },
    {
        'username': EMAIL,
        'success': True
    },
    {
        'password': PASSWORD + '_bad',
    },
    {
        'username': USERNAME + '_bad',
    },
    {
        'client_id': CLIENT_ID + '_bad'
    },
    {
        'username': EMAIL,
        'password': PASSWORD + '_bad',
    },
    {
        'client_secret': CLIENT_SECRET,
        'success': True
    },
    {
        'client_type': CONFIDENTIAL,
        'client_secret': CLIENT_SECRET,
        'success': True
    },
    {
        'client_secret': CLIENT_SECRET + '_bad',
        'success': True  # public clients should ignore the client_secret field
    },
    {
        'client_type': CONFIDENTIAL,
        'client_secret': CLIENT_SECRET + '_bad',
    },
]


@ddt.ddt
class AuthenticationTest(OAuth2TestCase):
    """
    Authentication test class.
    """
    def setUp(self):
        super(AuthenticationTest, self).setUp()
        user = self.user_factory.create(
            username=USERNAME,
            password=PASSWORD,
            email=EMAIL
        )
        self.set_user(user)

        self.auth_client = ClientFactory.create(
            client_id=CLIENT_ID,
            client_secret=CLIENT_SECRET
        )

        self.url = reverse('oauth2:access_token')

    @ddt.data(*AUTHENTICATION_TEST_DATA)
    def test_password_grants(self, data):
        self.auth_client.client_type = data.get('client_type', PUBLIC)
        self.auth_client.save()

        values = {
            'grant_type': 'password',
            'client_id': data.get('client_id', CLIENT_ID),
            'username': data.get('username', USERNAME),
            'password': data.get('password', PASSWORD),
        }

        client_secret = data.get('client_secret')
        if client_secret:
            values.update({'client_secret': client_secret})

        response = self.client.post(self.url, values)

        if data.get('success', False):
            self.assertEqual(200, response.status_code)
            self.assertIn('access_token', json.loads(response.content))
        else:
            self.assertEqual(400, response.status_code)



"""
Custom OAuth2 models
"""
from __future__ import absolute_import, division, print_function, unicode_literals

from django.db import models
from django.utils.encoding import python_2_unicode_compatible
from provider.oauth2.models import Client

# Import constants to force override of `provider.scope`
# See constants.py for explanation
from . import constants  # pylint: disable=unused-import


@python_2_unicode_compatible
class TrustedClient(models.Model):
    """
    By default `django-oauth2-provider` shows a consent form to the
    user after his credentials has been validated. Trusted clients
    bypass the user consent and redirect to the OAuth2 client
    directly.

    """
    client = models.ForeignKey(Client)

    class Meta(object):
        db_table = 'oauth2_provider_trustedclient'

    def __str__(self):
        return "{}".format(self.client)



"""
OAuth2 provider urls
"""
from __future__ import absolute_import, division, print_function, unicode_literals

from django.conf.urls import url
from django.contrib.auth.decorators import login_required
from django.views.decorators.csrf import csrf_exempt
from provider.oauth2.views import AccessTokenDetailView

from .views import AccessTokenView, Authorize, Capture, Redirect, UserInfoView

urlpatterns = [
    url(r'^authorize/?$', login_required(Capture.as_view()), name='capture'),
    url(r'^authorize/confirm/?$', login_required(Authorize.as_view()), name='authorize'),
    url(r'^redirect/?$', login_required(Redirect.as_view()), name='redirect'),
    url(r'^access_token/?$', csrf_exempt(AccessTokenView.as_view()), name='access_token'),
    url(r'^access_token/(?P<token>[\w]+)/$', csrf_exempt(AccessTokenDetailView.as_view()), name='access_token_detail'),
    url(r'^user_info/?$', csrf_exempt(UserInfoView.as_view()), name='user_info'),
]



"""
OAuth2 provider Django admin interface
"""
from __future__ import absolute_import, division, print_function, unicode_literals

from django.contrib import admin

from .models import TrustedClient


class TrustedClientAdmin(admin.ModelAdmin):
    "Django admin configuration for `TrustedClient` model"
    list_display = ('client',)


admin.site.register(TrustedClient, TrustedClientAdmin)



# pylint: disable=missing-docstring
from __future__ import absolute_import, division, print_function, unicode_literals

import datetime
import json

import jwt
import mock
from django.conf import settings
from django.test.utils import override_settings
from provider.oauth2.models import Grant
from provider.scope import check

from .. import constants, oidc
from ..oidc.core import IDToken
from .base import OAuth2TestCase
from .factories import AccessTokenFactory

BASE_DATETIME = datetime.datetime(1970, 1, 1)


@mock.patch('django.utils.timezone.now', mock.Mock(return_value=BASE_DATETIME))
@override_settings(OAUTH_OIDC_ISSUER='https://example.com/')
class BaseTestCase(OAuth2TestCase):
    def setUp(self):
        super(BaseTestCase, self).setUp()
        self.access_token = AccessTokenFactory(user=self.user, client=self.auth_client)


class IdTokenTest(BaseTestCase):
    def _get_actual_claims(self, access_token, nonce):
        with mock.patch('edx_oauth2_provider.oidc.handlers.datetime') as mock_datetime:
            mock_datetime.utcnow.return_value = BASE_DATETIME
            id_token = oidc.id_token(access_token, nonce)

            # Clear id token since a handler can change it.
            id_token.claims['sub'] = None

            return id_token.claims

    def _get_expected_claims(self, access_token, nonce):
        client = access_token.client

        # Basic OpenID Connect ID token claims
        expected = {
            'iss': settings.OAUTH_OIDC_ISSUER,
            'sub': None,
            'aud': client.client_id,
            'iat': 0,
            'exp': 30,
            'nonce': nonce
        }

        # Add profile scope claims
        if check(constants.PROFILE_SCOPE, access_token.scope):
            user = access_token.user
            expected.update({
                'name': user.get_full_name(),
                'given_name': user.first_name,
                'family_name': user.last_name,
                'preferred_username': user.username
            })

        # Add email scope claims
        if check(constants.EMAIL_SCOPE, access_token.scope):
            user = access_token.user
            expected.update({
                'email': user.email,
            })

        return expected

    def _decode(self, id_token, audience):
        return jwt.decode(id_token, 'test_secret', audience=audience, verify_expiration=False)

    def test_get_id_token(self):
        actual = self._get_actual_claims(self.access_token, self.nonce)
        expected = self._get_expected_claims(self.access_token, self.nonce)

        self.assertEqual(actual, expected)

        encoded = IDToken(self.access_token, scopes=[], claims=actual).encode('test_secret')
        self.assertEqual(self._decode(encoded, actual['aud']), expected)

    def test_get_id_token_with_profile(self):
        # Add the profile scope to access token
        self.access_token.scope |= constants.PROFILE_SCOPE
        self.access_token.save()

        claims = self._get_actual_claims(self.access_token, self.nonce)
        expected = self._get_expected_claims(self.access_token, self.nonce)
        self.assertEqual(claims, expected)

    def test_get_id_token_with_profile_and_email(self):
        # Add the profile scope to access token
        self.access_token.scope |= constants.PROFILE_SCOPE | constants.EMAIL_SCOPE
        self.access_token.save()

        claims = self._get_actual_claims(self.access_token, self.nonce)
        expected = self._get_expected_claims(self.access_token, self.nonce)
        self.assertEqual(claims, expected)

    def test_id_token_nonce(self):
        """ The nonce in the ID token should be the same nonce passed to the initial authorization call. """
        response = self.get_access_token_response(scope='openid')

        # Validate Grant has an associated nonce
        self.assertEqual(Grant.objects.filter(client=self.auth_client, user=self.user, nonce=self.nonce).count(), 1)

        # Validate ID token nonce
        response = json.loads(response.content)
        id_token = jwt.decode(response['id_token'], verify=False)
        self.assertEqual(id_token['nonce'], self.nonce)


class UserInfoTest(BaseTestCase):
    def assertIncludedClaims(self, claims, expected_scope=None, expected_claims=None):
        expected_scope = expected_scope if expected_scope else []
        expected_claims = expected_claims if expected_claims else []

        # Should always have a subject claim
        expected_claims.extend(['sub'])

        for scope in expected_scope:
            if scope == 'profile':
                expected_claims.extend(['preferred_username',
                                        'given_name',
                                        'family_name',
                                        'name'])
            if scope == 'email':
                expected_claims.extend(['email'])

        self.assertItemsEqual(claims.keys(), expected_claims)

    def test_values(self):
        # Add the profile scope to access token
        self.access_token.scope |= constants.PROFILE_SCOPE
        self.access_token.save()

        claims_request = {'userinfo': {'preferred_username': {'value': 'pedro'}}}
        claims = userinfo_claims(self.access_token, claims_request=claims_request)

        self.assertEqual(claims['preferred_username'], self.user.username)
        self.assertIn('sub', claims)
        self.assertEqual(len(claims), 2)  # should not return any more claims

    def test_not_recognized_values(self):
        id_token = oidc.userinfo(
            self.access_token,
            scope_request=['openid'],  # it should not matter if we don't add 'profile'
            claims_request={'userinfo': {'foo': {'value': 'bar'}}},
        )

        self.assertIn('sub', id_token.claims)
        self.assertEqual(len(id_token.claims), 1)

    def test_arguments(self):
        """ Test if the responses contain the requested claims according to permissions"""

        # TODO: replace with DDT test

        def userinfo_req(req):
            return {'userinfo': req}

        # Add the profile scope to access token
        self.access_token.scope |= constants.PROFILE_SCOPE
        self.access_token.save()

        claims = userinfo_claims(self.access_token)
        self.assertIncludedClaims(claims, ['profile'])

        claims = userinfo_claims(self.access_token, scope_request=['openid', 'profile', 'email'])
        self.assertIncludedClaims(claims, ['profile'])

        claims = userinfo_claims(self.access_token, scope_request=['openid', 'email'])
        self.assertIncludedClaims(claims)

        claims = userinfo_claims(
            self.access_token,
            claims_request=userinfo_req({'preferred_username': None})
        )
        self.assertIncludedClaims(claims, expected_claims=['preferred_username'])

        claims = userinfo_claims(self.access_token, scope_request=['email'])
        self.assertIncludedClaims(claims)

        claims = userinfo_claims(
            self.access_token,
            scope_request=['profile'],
            claims_request=userinfo_req({'preferred_username': None})
        )
        self.assertIncludedClaims(claims, ['profile'])

        claims = userinfo_claims(
            self.access_token, scope_request=['email'],
            claims_request=userinfo_req({'preferred_username': None})
        )
        self.assertIncludedClaims(claims, expected_claims=['preferred_username'])

        claims = userinfo_claims(
            self.access_token,
            claims_request=userinfo_req({'email': None, 'preferred_username': None})
        )
        self.assertIncludedClaims(claims, expected_claims=['preferred_username'])

        # Check with extra scopes

        self.access_token.scope |= constants.PROFILE_SCOPE | constants.EMAIL_SCOPE
        self.access_token.save()

        claims = userinfo_claims(
            self.access_token,
            claims_request=userinfo_req({'email': None, 'preferred_username': None})
        )
        self.assertIncludedClaims(claims, expected_claims=['email', 'preferred_username'])

        claims = userinfo_claims(
            self.access_token,
            scope_request=['email'],
            claims_request=userinfo_req({'preferred_username': None})
        )
        self.assertIncludedClaims(claims, ['email'], ['preferred_username'])

        claims = userinfo_claims(self.access_token)
        self.assertIncludedClaims(claims, ['profile', 'email'])


def userinfo_claims(access_token, scope_request=None, claims_request=None):
    id_token = oidc.userinfo(access_token, scope_request, claims_request)
    return id_token.claims



"""
OAuth2 provider `django-oauth2-provider` authentication backends

"""
from __future__ import absolute_import, division, print_function, unicode_literals

from .forms import PublicPasswordGrantForm


class PublicPasswordBackend(object):
    """
    Simple client authentication wrapper backends that delegates to
    `oauth2_provider.forms.PublicPasswordGrantForm`

    """

    def authenticate(self, request=None):
        """ Returns client if correctly authenticated. Otherwise returns None """

        if request is None:
            return None

        form = PublicPasswordGrantForm(request.POST)

        # pylint: disable=no-member
        if form.is_valid():
            return form.cleaned_data.get('client')

        return None



"""
OAuth2 and OpenID Connect settings

"""
from __future__ import absolute_import, division, print_function, unicode_literals

import provider.constants
import provider.oauth2.forms
import provider.scope
from django.conf import settings

# REQUIRED Issuer Identifider (the `iss` field in the `id_token`)
OAUTH_OIDC_ISSUER = settings.OAUTH_OIDC_ISSUER

# OAUTH/2 OpenID Connect scopes

# Use bit-shifting so that scopes can be easily combined and checked.
DEFAULT_SCOPE = 0
OPEN_ID_SCOPE = 1 << 0
PROFILE_SCOPE = 1 << 1
EMAIL_SCOPE = 1 << 2
COURSE_STAFF_SCOPE = 1 << 3
COURSE_INSTRUCTOR_SCOPE = 1 << 4
PERMISSIONS = 1 << 5

# Scope setting as required by django-oauth2-provider
# The default scope value is SCOPES[0][0], which in this case is zero.
# `django-oauth2-provider` considers a scope value of zero as empty,
# ignoring its name when requested.
SCOPES = (
    (DEFAULT_SCOPE, 'default'),
    (OPEN_ID_SCOPE, 'openid'),
    (PROFILE_SCOPE, 'profile'),
    (EMAIL_SCOPE, 'email'),
    (COURSE_STAFF_SCOPE, 'course_staff'),
    (COURSE_INSTRUCTOR_SCOPE, 'course_instructor'),
    (PERMISSIONS, 'permissions')
)

SCOPE_NAMES = [(name, name) for (value, name) in SCOPES]
SCOPE_NAME_DICT = dict([(name, value) for (value, name) in SCOPES])
SCOPE_VALUE_DICT = dict([(value, name) for (value, name) in SCOPES])


# OpenID Connect claim handlers

DEFAULT_ID_TOKEN_HANDLERS = (
    'oauth2_provider.oidc.handlers.BasicIDTokenHandler',
    'oauth2_provider.oidc.handlers.ProfileHandler',
    'oauth2_provider.oidc.handlers.EmailHandler',
)

DEFAULT_USERINFO_HANDLERS = (
    'oauth2_provider.oidc.handlers.BasicUserInfoHandler',
    'oauth2_provider.oidc.handlers.ProfileHandler',
    'oauth2_provider.oidc.handlers.EmailHandler',
)

ID_TOKEN_HANDLERS = getattr(settings, 'OAUTH_OIDC_ID_TOKEN_HANDLERS', DEFAULT_ID_TOKEN_HANDLERS)
USERINFO_HANDLERS = getattr(settings, 'OAUTH_OIDC_USERINFO_HANDLERS', DEFAULT_USERINFO_HANDLERS)


# Override django-oauth2-provider scopes (OAUTH_SCOPES)
#
# `provider.scopes` values are loaded from Django settings. However,
# we don't want to rely on customizable settings, and instead use the
# set of defaults above. One problem, is that in many places
# `django-oauth2-provider` creates module globals with the values from
# settings, making them hard to override.
#
# TODO: This solution is bit ugly, but viable for now. A better fix is
# to make SCOPES be lazy evaluated at the django-oauth2-provider level
# using django.utils.functional.lazy

provider.constants.SCOPES = SCOPES
provider.constants.DEFAULT_SCOPES = SCOPES

provider.scope.SCOPES = SCOPES
provider.scope.SCOPE_NAMES = SCOPE_NAMES
provider.scope.SCOPE_NAME_DICT = SCOPE_NAME_DICT
provider.scope.SCOPE_VALUE_DICT = SCOPE_VALUE_DICT

provider.oauth2.forms.SCOPES = SCOPES
provider.oauth2.forms.SCOPE_NAMES = SCOPE_NAMES

AUTHORIZED_CLIENTS_SESSION_KEY = getattr(settings, 'OAUTH_OIDC_AUTHORIZED_CLIENTS_SESSION_KEY', 'authorized_clients')



"""
Functions to collect OpenID Connect values from claim handlers.

For details on the format of the claim handlers, see
:mod:`oauth2_provider.oicd.handlers`

None: The functions in this module assume the `openid` scope is implied.
"""
# pylint: disable=missing-docstring
from __future__ import absolute_import, division, print_function, unicode_literals

import provider.scope

REQUIRED_SCOPES = ['openid']

CLAIM_REQUEST_FIELDS = ['value', 'values', 'essential']


def collect(handlers, access_token, scope_request=None, claims_request=None):
    """
    Collect all the claims values from the `handlers`.

    Arguments:
      handlers (list): List of claim :class:`Handler` classes.
      access_token (:class:AccessToken): Associated access token.
      scope_request (list): List of requested scopes.
      claims_request (dict): Dictionary with only the relevant section of a
          OpenID Connect claims request.

    Returns a list of the scopes from `scope_request` that are authorized, and a
    dictionary of the claims associated with the authorized scopes in
    `scope_request`, and additionally, the authorized claims listed in
    `claims_request`.

    """
    user = access_token.user
    client = access_token.client

    # Instantiate handlers. Each handler is instanciated only once, allowing the
    # handler to keep state in-between calls to its scope and claim methods.

    handlers = [cls() for cls in handlers]

    # Find all authorized scopes by including the access_token scopes.  Note
    # that the handlers determine if a scope is authorized, not its presense in
    # the access_token.

    required_scopes = set(REQUIRED_SCOPES)
    token_scopes = set(provider.scope.to_names(access_token.scope))
    authorized_scopes = _collect_scopes(handlers, required_scopes | token_scopes, user, client)

    # Select only the authorized scopes from the requested scopes.

    scope_request = set(scope_request) if scope_request else set()
    scopes = required_scopes | (authorized_scopes & scope_request)

    # Find all authorized claims names for the authorized_scopes.

    authorized_names = _collect_names(handlers, authorized_scopes, user, client)

    # Select only the requested claims if no scope has been requested. Selecting
    # scopes has prevalence over selecting claims.

    claims_request = _validate_claim_request(claims_request)

    # Add the requested claims that are authorized to the response.

    requested_names = set(claims_request.keys()) & authorized_names
    names = _collect_names(handlers, scopes, user, client) | requested_names

    # Get the values for the claims.

    claims = _collect_values(
        handlers,
        names=names,
        user=user,
        client=client,
        values=claims_request or {}
    )

    return authorized_scopes, claims


def _collect_scopes(handlers, scopes, user, client):
    """ Get a set of all the authorized scopes according to the handlers. """
    results = set()

    data = {'user': user, 'client': client}

    def visitor(scope_name, func):
        claim_names = func(data)
        # If the claim_names is None, it means that the scope is not authorized.
        if claim_names is not None:
            results.add(scope_name)

    _visit_handlers(handlers, visitor, 'scope', scopes)

    return results


def _collect_names(handlers, scopes, user, client):
    """ Get the names of the claims supported by the handlers for the requested scope. """

    results = set()

    data = {'user': user, 'client': client}

    def visitor(_scope_name, func):
        claim_names = func(data)
        # If the claim_names is None, it means that the scope is not authorized.
        if claim_names is not None:
            results.update(claim_names)

    _visit_handlers(handlers, visitor, 'scope', scopes)

    return results


def _collect_values(handlers, names, user, client, values):
    """ Get the values from the handlers of the requested claims. """

    results = {}

    def visitor(claim_name, func):
        data = {'user': user, 'client': client}
        data.update(values.get(claim_name) or {})
        claim_value = func(data)
        # If the claim_value is None, it means that the claim is not authorized.
        if claim_value is not None:
            # New values overwrite previous results
            results[claim_name] = claim_value

    _visit_handlers(handlers, visitor, 'claim', names)

    return results


def _validate_claim_request(claims, ignore_errors=False):
    """
    Validates a claim request section (`userinfo` or `id_token`) according
    to section 5.5 of the OpenID Connect specification:

    - http://openid.net/specs/openid-connect-core-1_0.html#ClaimsParameter

    Returns a copy of the claim request with only the valid fields and values.

    Raises ValueError is the claim request is invalid and `ignore_errors` is False

    """

    results = {}
    claims = claims if claims else {}

    for name, value in claims.iteritems():
        if value is None:
            results[name] = None
        elif isinstance(value, dict):
            results[name] = _validate_claim_values(name, value, ignore_errors)
        else:
            if not ignore_errors:
                msg = 'Invalid claim {}.'.format(name)
                raise ValueError(msg)

    return results


def _validate_claim_values(name, value, ignore_errors):
    """ Helper for `validate_claim_request` """
    results = {'essential': False}
    for key, value in value.iteritems():
        if key in CLAIM_REQUEST_FIELDS:
            results[key] = value
        else:
            if not ignore_errors:
                msg = 'Unknown attribute {} in claim value {}.'.format(key, name)
                raise ValueError(msg)
    return results


def _visit_handlers(handlers, visitor, prefix, suffixes):
    """ Use visitor partern to collect information from handlers """

    results = []
    for handler in handlers:
        for suffix in suffixes:
            func = getattr(handler, '{}_{}'.format(prefix, suffix).lower(), None)
            if func:
                results.append(visitor(suffix, func))

    return results



"""
Trusted client tests.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

from django.core.urlresolvers import reverse

from .base import OAuth2TestCase
from .util import normpath


class TrustedClientTest(OAuth2TestCase):
    """
    Trusted client test class.
    """
    def test_trusted_client(self):
        response = self.login_and_authorize(trusted=True)

        # Check for valid redirect
        self.assertEqual(302, response.status_code)
        self.assertEqual(reverse('oauth2:redirect'), normpath(response['Location']))

    def test_untrusted_client(self):
        response = self.login_and_authorize(trusted=False, validate_session=False)

        # Check if consent form is being shown
        form_action = 'action="{}"'.format(normpath(reverse("oauth2:authorize")))
        self.assertContains(response, form_action, status_code=200)



"""
Management command used to create an OAuth2 client in the database.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import json

from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError
from django.core.management.base import BaseCommand, CommandError
from django.core.validators import URLValidator
from provider.constants import CONFIDENTIAL, PUBLIC
from provider.oauth2.models import Client

from ...models import TrustedClient

ARG_STRING = '<url> <redirect_uri> <client_type: "confidential" | "public">'


class Command(BaseCommand):
    """
    create_oauth2_client command class
    """
    help = 'Create a new OAuth2 Client. Outputs a serialized representation of the newly-created Client.'

    def add_arguments(self, parser):
        super(Command, self).add_arguments(parser)

        # Required positional arguments.
        parser.add_argument(
            'url',
            help="Url."
        )
        parser.add_argument(
            'redirect_uri',
            help="Redirect URI."
        )
        parser.add_argument(
            'client_type',
            help="Client type."
        )

        # Optional options.
        parser.add_argument(
            '-u',
            '--username',
            help="Username of a user to associate with the Client."
        )
        parser.add_argument(
            '-n',
            '--client_name',
            help="String to assign as the Client name."
        )
        parser.add_argument(
            '-i',
            '--client_id',
            help="String to assign as the Client ID."
        )
        parser.add_argument(
            '-s',
            '--client_secret',
            help="String to assign as the Client secret. Should not be shared."
        )
        parser.add_argument(
            '-t',
            '--trusted',
            action='store_true',
            help="Designate the Client as trusted. Trusted Clients bypass the user consent "
                 "form typically displayed after validating the user's credentials."
        )
        parser.add_argument(
            '--logout_uri',
            help="Client logout URI. This value will be used for single sign out."
        )

    def handle(self, *args, **options):
        self._clean_required_args(options['url'], options['redirect_uri'], options['client_type'])
        self._parse_options(options)

        client_id = self.fields.get('client_id')
        trusted = self.fields.pop('trusted')

        # Check if client ID is already in use. If so, fetch existing Client and update fields.
        client_id_claimed = Client.objects.filter(client_id=client_id).exists()
        if client_id_claimed:
            client = Client.objects.get(client_id=client_id)

            for key, value in self.fields.items():
                setattr(client, key, value)

            client.save()
        else:
            client = Client.objects.create(**self.fields)

        if trusted:
            TrustedClient.objects.get_or_create(client=client)
        else:
            try:
                TrustedClient.objects.get(client=client).delete()
            except TrustedClient.DoesNotExist:
                pass

        serialized = json.dumps(client.serialize(), indent=4)
        self.stdout.write(serialized)

    def _clean_required_args(self, url, redirect_uri, client_type):
        """
        Validate and clean the command's arguments.

        Arguments:
            url (str): Client's application URL.
            redirect_uri (str): Client application's OAuth2 callback URI.
            client_type (str): Client's type, indicating whether the Client application
                is capable of maintaining the confidentiality of its credentials (e.g., running on a
                secure server) or is incapable of doing so (e.g., running in a browser).

        Raises:
            CommandError, if the URLs provided are invalid, or if the client type provided is invalid.
        """
        # Validate URLs
        for url_to_validate in (url, redirect_uri):
            try:
                URLValidator()(url_to_validate)
            except ValidationError:
                raise CommandError("URLs provided are invalid. Please provide valid application and redirect URLs.")

        # Validate and map client type to the appropriate django-oauth2-provider constant
        client_type = client_type.lower()
        client_type = {
            'confidential': CONFIDENTIAL,
            'public': PUBLIC
        }.get(client_type)

        if client_type is None:
            raise CommandError("Client type provided is invalid. Please use one of 'confidential' or 'public'.")

        self.fields = {  # pylint: disable=attribute-defined-outside-init
            'url': url,
            'redirect_uri': redirect_uri,
            'client_type': client_type,
        }

    def _parse_options(self, options):
        """Parse the command's options.

        Arguments:
            options (dict): Options with which the command was called.

        Raises:
            CommandError, if a user matching the provided username does not exist.
        """
        for key in ('username', 'client_name', 'client_id', 'client_secret', 'trusted', 'logout_uri'):
            value = options.get(key)
            if value is not None:
                self.fields[key] = value

        username = self.fields.pop('username', None)
        if username is not None:
            try:
                user_model = get_user_model()
                self.fields['user'] = user_model.objects.get(username=username)
            except user_model.DoesNotExist:
                raise CommandError("User matching the provided username does not exist.")

        # The keyword argument 'name' conflicts with that of `call_command()`. We instead
        # use 'client_name' up to this point, then swap it out for the expected field, 'name'.
        client_name = self.fields.pop('client_name', None)
        if client_name is not None:
            self.fields['name'] = client_name

        logout_uri = self.fields.get('logout_uri')

        if logout_uri:
            try:
                URLValidator()(logout_uri)
            except ValidationError:
                raise CommandError("The logout_uri is invalid.")



# pylint: disable=missing-docstring
from __future__ import absolute_import, division, print_function, unicode_literals

import json
import uuid
from urlparse import urlparse

import jwt
import provider.scope
from django.core.urlresolvers import reverse
from django.http import QueryDict
from django.test import TestCase

from ..constants import AUTHORIZED_CLIENTS_SESSION_KEY
from ..models import TrustedClient
from .factories import AccessTokenFactory, ClientFactory, TrustedClientFactory, UserFactory
from .util import normpath


class BaseTestCase(TestCase):
    def setUp(self):
        super(BaseTestCase, self).setUp()

        self.client_secret = 'some_secret'
        self.auth_client = ClientFactory(client_secret=self.client_secret)

        self.password = 'some_password'
        self.user_factory = UserFactory

        self.user = None
        self.access_token = None
        self.set_user(self.make_user())

    def make_user(self):
        return self.user_factory(password=self.password)

    def set_user(self, user):
        self.user = user

    def set_trusted(self, client, trusted=True):
        if trusted:
            TrustedClientFactory.create(client=client)
        else:
            TrustedClient.objects.filter(client=client).delete()


class OAuth2TestCase(BaseTestCase):
    def setUp(self):
        super(OAuth2TestCase, self).setUp()
        self.nonce = str(uuid.uuid4())

    def login_and_authorize(self, scope=None, claims=None, trusted=False, validate_session=True):
        """ Login into client using OAuth2 authorization flow. """

        self.set_trusted(self.auth_client, trusted)
        self.client.login(username=self.user.username, password=self.password)

        client_id = self.auth_client.client_id
        payload = {
            'client_id': client_id,
            'redirect_uri': self.auth_client.redirect_uri,
            'response_type': 'code',
            'state': 'some_state',
            'nonce': self.nonce,
        }
        _add_values(payload, 'id_token', scope, claims)

        response = self.client.get(reverse('oauth2:capture'), payload)
        self.assertEqual(302, response.status_code)

        response = self.client.get(reverse('oauth2:authorize'), payload)

        if validate_session:
            self.assertListEqual(self.client.session[AUTHORIZED_CLIENTS_SESSION_KEY], [client_id])

        return response

    def get_access_token_response(self, scope=None, claims=None):
        """ Get a new access token using the OAuth2 authorization flow. """
        response = self.login_and_authorize(scope, claims, trusted=True)
        self.assertEqual(302, response.status_code)
        self.assertEqual(reverse('oauth2:redirect'), normpath(response['Location']))

        response = self.client.get(reverse('oauth2:redirect'))
        self.assertEqual(302, response.status_code)

        query = QueryDict(urlparse(response['Location']).query)

        payload = {
            'grant_type': 'authorization_code',
            'client_id': self.auth_client.client_id,
            'client_secret': self.client_secret,
            'code': query['code'],
        }
        _add_values(payload, 'id_token', scope, claims)

        response = self.client.post(reverse('oauth2:access_token'), payload)
        return response


class IDTokenTestCase(OAuth2TestCase):
    def get_id_token_values(self, scope=None, claims=None):
        """ Get a new id_token using the OIDC authorization flow. """

        self.assertIn('openid', scope.split())

        response = self.get_access_token_response(scope, claims)
        self.assertEqual(response.status_code, 200)

        values = json.loads(response.content)
        self.assertIn('access_token', values)

        id_token = values['id_token']
        secret = self.auth_client.client_secret
        audience = self.auth_client.client_id
        self.assertValidIDToken(id_token, secret, audience)

        scopes = values['scope'].split()
        claims = self.parse_id_token(id_token)

        # Should always be included
        self.assertIn('iss', claims)
        self.assertIn('sub', claims)

        return scopes, claims

    def parse_id_token(self, id_token):
        claims = jwt.decode(id_token, verify=False)
        return claims

    def assertValidIDToken(self, id_token, secret, audience):
        try:
            jwt.decode(id_token, secret, audience=audience)
        except jwt.DecodeError:
            assert False


class UserInfoTestCase(BaseTestCase):
    def setUp(self):
        super(UserInfoTestCase, self).setUp()
        self.path = reverse('oauth2:user_info')
        self.set_user(self.user)

    def set_user(self, user):
        super(UserInfoTestCase, self).set_user(user)
        self.access_token = AccessTokenFactory(user=self.user, client=self.auth_client)

    def set_access_token_scope(self, scope):
        self.access_token.scope = provider.scope.to_int(*scope.split())
        self.access_token.save()  # pylint: disable=no-member

    def get_with_authorization(self, path, access_token=None, payload=None):
        kwargs = {}
        if access_token:
            kwargs['HTTP_AUTHORIZATION'] = 'Bearer %s' % access_token

        return self.client.get(path, payload, **kwargs)

    def get_userinfo(self, token=None, scope=None, claims=None):
        payload = _add_values({}, 'userinfo', scope, claims)
        response = self.get_with_authorization(self.path, token, payload)
        values = json.loads(response.content)
        return response, values


def _add_values(data, endpoint, scope=None, claims=None):
    if scope:
        data['scope'] = scope
    if claims:
        data['claims'] = json.dumps({endpoint: claims})
    return data



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('oauth2', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='TrustedClient',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('client', models.ForeignKey(to='oauth2.Client')),
            ],
            options={
                'db_table': 'oauth2_provider_trustedclient',
            },
        ),
    ]



"""
Util methods for testing.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import os.path
import urlparse


def normpath(url):
    """Get the normalized path of a URL"""
    return os.path.normpath(urlparse.urlparse(url).path)



# pylint: disable=missing-docstring
from __future__ import absolute_import, division, print_function, unicode_literals

from django.contrib.auth.models import User
import factory
from factory.django import DjangoModelFactory
import provider.oauth2.models
from provider.constants import CONFIDENTIAL

from .. import models


class UserFactory(DjangoModelFactory):
    class Meta(object):
        model = User
        django_get_or_create = ('username', )

    username = factory.Sequence(u'robot_{0}'.format)
    email = factory.Sequence(u'robot_{0}@edx.org'.format)
    password = factory.PostGenerationMethodCall('set_password', 'some_password')

    first_name = factory.Sequence(u'Robot{0}'.format)
    last_name = 'Test'

    is_staff = False
    is_active = True
    is_superuser = False


class ClientFactory(DjangoModelFactory):
    class Meta(object):
        model = provider.oauth2.models.Client

    client_id = factory.Sequence(u'client_{0}'.format)
    client_secret = 'some_secret'
    client_type = CONFIDENTIAL

    url = 'http://example.com'
    redirect_uri = 'http://example.com/application'


class TrustedClientFactory(DjangoModelFactory):
    class Meta(object):
        model = models.TrustedClient


class AccessTokenFactory(DjangoModelFactory):
    class Meta(object):
        model = provider.oauth2.models.AccessToken
        django_get_or_create = ('user', 'client')


class RefreshTokenFactory(DjangoModelFactory):
    class Meta(object):
        model = provider.oauth2.models.RefreshToken
        django_get_or_create = ('user', 'client')



#!/usr/bin/env python

from setuptools import find_packages, setup

import edx_oauth2_provider

setup(
    name='edx-oauth2-provider',
    version=edx_oauth2_provider.__version__,
    description='Provide OAuth2 access to edX installations',
    author='edX',
    url='https://github.com/edx/edx-oauth2-provider',
    license='AGPL',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Affero General Public License v3',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Framework :: Django',
    ],
    packages=find_packages(exclude=['tests']),
    install_requires=[
        'django>=1.8.7,<2.0',
        'edx-django-oauth2-provider>=1.2.1,<2.0.0',
        'PyJWT>=1.4.0,<2.0.0'
    ]
)



"""
OpenID Connect core related utility functions.

Defines utility functions to process the ID Token and UserInfo
endpoints according to the OpenID Connect specification.

"""
from __future__ import absolute_import, division, print_function, unicode_literals

import jwt
import provider.scope
from django.utils.module_loading import import_string

from .. import constants
from .collect import collect

HANDLERS = {
    'id_token': [import_string(cls) for cls in constants.ID_TOKEN_HANDLERS],
    'userinfo': [import_string(cls) for cls in constants.USERINFO_HANDLERS]
}


class IDToken(object):
    """
    Simple container for OpenID Connect related responses.

    Attributes:
        access_token (:class:`AccessToken`): Associated Access Token object.
        scopes (list): List of scope names.
        claims (dict): Dictionary of claim names and values.

    """

    def __init__(self, access_token, scopes, claims):
        self.access_token = access_token
        self.scopes = scopes
        self.claims = claims

    def encode(self, secret, algorithm='HS256'):
        """
        Encode the set of claims to the JWT (JSON Web Token) format
        according to the OpenID Connect specification:

        http://openid.net/specs/openid-connect-basic-1_0.html#IDToken

        Arguments:
            claims (dict): A dictionary with the OpenID Connect claims.
            secret (str): Secret used to encode the id_token.
            algorithm (str): Algorithm used for encoding.
                Defaults to HS256.

        Returns encoded JWT token string.

        """

        return jwt.encode(self.claims, secret, algorithm)


def id_token(access_token, nonce=None, claims_request=None):
    """
    Returns data required for an OpenID Connect ID Token according to:

    - http://openid.net/specs/openid-connect-basic-1_0.html#IDToken

    Arguments:
        access_token (:class:`AccessToken`): Associated OAuth2 access token.
        nonce (str): Optional nonce to protect against replay attacks.
        claims_request (dict): Optional dictionary with the claims request parameters.

    Information on the `claims_request` parameter specification:

    - http://openid.net/specs/openid-connect-core-1_0.html#ClaimsParameter

    Returns an :class:`IDToken` instance with the scopes from the
    access_token and the corresponding claims. Claims in the
    `claims_request` paramater id_token section will be included *in
    addition* to the ones corresponding to the scopes specified in the
    `access_token`.

    """

    handlers = HANDLERS['id_token']

    # Select only the relevant section of the claims request.
    claims_request_section = claims_request.get('id_token', {}) if claims_request else {}

    scope_request = provider.scope.to_names(access_token.scope)

    if nonce:
        claims_request_section.update({'nonce': {'value': nonce}})

    scopes, claims = collect(
        handlers,
        access_token,
        scope_request=scope_request,
        claims_request=claims_request_section,
    )

    return IDToken(access_token, scopes, claims)


def userinfo(access_token, scope_request=None, claims_request=None):
    """
    Returns data required for an OpenID Connect UserInfo response, according to:

    http://openid.net/specs/openid-connect-basic-1_0.html#UserInfoResponse

    Supports scope and claims request parameter as described in:

    - http://openid.net/specs/openid-connect-core-1_0.html#ScopeClaims
    - http://openid.net/specs/openid-connect-core-1_0.html#ClaimsParameter

    Arguments: access_token (:class:`AccessToken`): Associated access
        token.  scope_request (list): Optional list of requested
        scopes. Only scopes authorized in the `access_token` will be
            considered.  claims_request
        (dict): Optional dictionary with a claims request parameter.

    Information on the claims request parameter specification:

    - http://openid.net/specs/openid-connect-core-1_0.html#ClaimsParameter

    As a convinience, if neither `scope_request` or user_info claim is
    specified in the `claims_request`, it will return the claims for
    all the scopes in the `access_token`.

    Returns an :class:`IDToken` instance with the scopes from the
    `scope_request` and the corresponding claims. Claims in the
    `claims_request` paramater userinfo section will be included *in
    addition* to the ones corresponding to `scope_request`.

    """

    handlers = HANDLERS['userinfo']

    # Select only the relevant section of the claims request.
    claims_request_section = claims_request.get('userinfo', {}) if claims_request else {}

    # If nothing is requested, return the claims for the scopes in the access token.
    if not scope_request and not claims_request_section:
        scope_request = provider.scope.to_names(access_token.scope)
    else:
        scope_request = scope_request

    scopes, claims = collect(
        handlers,
        access_token,
        scope_request=scope_request,
        claims_request=claims_request_section,
    )

    return IDToken(access_token, scopes, claims)



"""
ID token tests
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import json

from django.test.utils import override_settings

from .base import IDTokenTestCase

ISSUER = 'https://example.com/oauth2'


@override_settings(OAUTH_OIDC_ISSUER=ISSUER)
class OIDCScopeTest(IDTokenTestCase):
    """Test OpenID Connect ID Token scopes"""

    def test_default_scope(self):
        response = self.get_access_token_response()
        self.assertEqual(200, response.status_code)

        values = json.loads(response.content)
        scope = values.get('scope', '').split()

        # The default scope should be empty
        self.assertEqual(len(scope), 0)

    def test_id_token(self):
        scopes, claims = self.get_id_token_values('openid')

        self.assertIn('openid', scopes)

        required_claims = ['iss', 'iat', 'sub', 'exp', 'aud']
        for name in required_claims:
            self.assertIn(name, claims)

        self.assertEqual(ISSUER, claims['iss'])

    def test_sub_claim(self):
        # Verify that the 'sub' claim is unique for each user
        user_a = self.user
        _scope, claims_a = self.get_id_token_values('openid')

        self.set_user(self.make_user())
        _scope, claims_b = self.get_id_token_values('openid')
        self.assertNotEqual(claims_a['sub'], claims_b['sub'])

        # generate another token for user_a
        self.set_user(user_a)
        _scope, claims_c = self.get_id_token_values('openid')
        self.assertEqual(claims_a['sub'], claims_c['sub'])

    def test_profile_scope(self):
        scopes, claims = self.get_id_token_values('openid profile')

        self.assertIn('openid', scopes)
        self.assertIn('profile', scopes)

        self.assertIn('preferred_username', claims)

    def test_email_scope(self):
        scopes, claims = self.get_id_token_values('openid email')

        self.assertIn('openid', scopes)
        self.assertIn('email', scopes)

        self.assertIn('email', claims)

    def test_claims_request(self):
        scope_request = 'openid profile'

        _, claims = self.get_id_token_values(scope_request)

        self.assertNotIn('test', claims)

        claims_request = {
            'test': {
                'essential': True,
                'values': [2, 6, 8, 12]
            }
        }

        _, claims = self.get_id_token_values(scope_request, claims_request)

        self.assertIn('test', claims)
        self.assertListEqual([2, 6, 8], sorted(claims['test']))



"""
Userinfo tests.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

from .base import UserInfoTestCase


class UserInfoViewTest(UserInfoTestCase):
    """
    User info view tests.
    """
    def test_authorized(self):
        token = self.access_token.token
        self.set_access_token_scope('openid')

        response, _ = self.get_userinfo(token)
        self.assertEqual(response.status_code, 200)

    def test_unauthorized(self):
        # No token
        response, values = self.get_userinfo()
        self.assertEqual(response.status_code, 401)
        self.assertEqual(values['error'], 'access_denied')

        # Bad Token
        token = 'bad_token'
        response, _ = self.get_userinfo(token)
        self.assertEqual(response.status_code, 401)
        self.assertEqual(values['error'], 'access_denied')

        # No openid scope in access token
        token = self.access_token.token
        self.set_access_token_scope('something')

        response, _ = self.get_userinfo(token)
        self.assertEqual(response.status_code, 400)
        self.assertEqual(values['error'], 'access_denied')

    def test_scope_request(self):
        scope = 'openid profile'
        self.set_access_token_scope(scope)

        token = self.access_token.token
        response, claims = self.get_userinfo(token, scope)
        self.assertEqual(response.status_code, 200)

        user = self.user
        expected = {
            u'preferred_username': user.username,
            u'given_name': user.first_name,
            u'family_name': user.last_name,
            u'name': user.get_full_name(),
        }

        for key, value in expected.iteritems():
            self.assertEqual(claims[key], value)

    def test_claims_request(self):
        valid_scope = 'openid profile'
        self.set_access_token_scope(valid_scope)

        token = self.access_token.token

        scope_request = 'openid whatever'
        claims_request = {
            'name': None,
            'foo': {'value': 'one'},
            'what': {'values': ['one', 'two']},
        }

        response, claims = self.get_userinfo(token, scope_request, claims_request)

        self.assertEqual(response.status_code, 200)
        self.assertIn('sub', claims)
        self.assertIn('name', claims)
        self.assertEqual(len(claims), 2)

        scope_request = 'openid profile'
        claims_request = {
            'test': {'essential': True}
        }

        response, claims = self.get_userinfo(token, scope_request, claims_request)

        self.assertIn('test', claims)



# pylint: disable=missing-docstring
from __future__ import absolute_import, division, print_function, unicode_literals


class DummyHandler(object):
    def scope_profile(self, data):  # pylint: disable=unused-argument
        return ['test']

    def claim_test(self, data):
        """
        Test claim.

        Only has a value if the claim request has essential=True. The
        value is range(10), or the intersection with the values in the
        claims request if any.

        """
        if data.get('essential'):
            values = range(10)
            if data.get('values'):
                values = list(set(values) & set(data.get('values')))
            return values
        return None
